"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.offset = offset;
exports.single = single;
exports.double = double;
exports.unsafe = unsafe;

var _bytes = require("@capnp-js/bytes");

var _readData = require("@capnp-js/read-data");

var _tinyUint = require("@capnp-js/tiny-uint");

function offset(ref) {
  const half = (0, _readData.int32)(ref.segment.raw, ref.position) & 0xfffffffc;
  return half + half;
}

function single(far, typeBits) {
  return {
    typeBits,
    hi: (0, _readData.int32)(far.segment.raw, far.position + 4),
    object: {
      segment: far.segment,
      position: far.position + 8 + offset(far)
    }
  };
}

function double(table, far, typeBits) {
  const object = {
    segment: table.segment((0, _readData.int32)(far.segment.raw, far.position + 4) >>> 0),
    position: offset(far)
  };
  const hi = (0, _readData.int32)(far.segment.raw, far.position + 12);
  return {
    typeBits,
    hi,
    object
  };
}

function unsafe(table, ref) {
  const lsb = (0, _bytes.get)(ref.position, ref.segment.raw);
  const typeBits = (0, _tinyUint.u2_mask)(lsb, 0x03);

  if (typeBits === 0x02) {
    const far = {
      segment: table.segment((0, _readData.int32)(ref.segment.raw, ref.position + 4) >>> 0),
      position: ((0, _readData.int32)(ref.segment.raw, ref.position) & 0xfffffff8) >>> 0
    };
    /* By using the 0x01 mask on the object's type bits, capability "objects"
     * get converted to nonsense lists. */

    if ((0, _tinyUint.u3_mask)(lsb, 0x04) === 0x00) {
      /* Single Hop */
      return single(far, (0, _tinyUint.u2_mask)((0, _bytes.get)(far.position, far.segment.raw), 0x01));
    } else {
      /* Double Hop */
      return double(table, far, (0, _tinyUint.u2_mask)((0, _bytes.get)(far.position + 8, far.segment.raw), 0x01));
    }
  } else {
    /* `ref` can be reconstructed as a `Word<S>` by using
     * `table.segment(ref.segment.id)`. The covariance of the `segment`
     * property on `Word<.>` makes these `ref`-to-`Word<S>` safe type casts. An
     * `S`-typed version of `ref`'s segment can be obtained with
     * `table.segment(ref.segment.id)`, so `ref.segment`-to-`S` is also a safe
     * type cast. */
    return {
      typeBits,
      hi: (0, _readData.int32)(ref.segment.raw, ref.position + 4),
      object: typeBits === 0x03 ? ref : {
        // eslint-disable-line flowtype/no-weak-types
        segment: ref.segment,
        // eslint-disable-line flowtype/no-weak-types
        position: ref.position + 8 + offset(ref)
      }
    };
  }
}